<?php

namespace Drupal\commerce_easytransac\PluginForm\EasyTransac;

use Drupal\commerce_easytransac\EasyTransac;
use Drupal\commerce_easytransac\Modifier;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PluginForm\PaymentGatewayFormBase;
use Drupal\Core\Form\FormStateInterface;
use EasyTransac\Entities\PaymentPageTransaction;
use EasyTransac\Requests\PaymentPage;

/**
 * EasyTransac payment gateway plugin form to add a payment using a payment page.
 */
class PaymentAddForm extends PaymentGatewayFormBase {

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $this->entity;
    $order = $payment->getOrder();
    if (!$order) {
      throw new \InvalidArgumentException('Payment entity with no order reference given to PaymentAddForm.');
    }

    // The payment amount should not exceed the remaining order balance.
    $balance = $order->getBalance();
    if (!$balance) {
      throw new \InvalidArgumentException('The payment amount cannot be determined due to a NULL order balance.');
    }
    $amount = $balance->isPositive() ? $balance : $balance->multiply(0);

    $form['amount'] = [
      '#type' => 'commerce_price',
      '#title' => $this->t('Amount'),
      '#default_value' => $amount->toArray(),
      '#required' => TRUE,
    ];

    $form['capture'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Capture the transaction'),
      '#default_value' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValue($form['#parents']);

    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $this->entity;
    $payment->amount = $values['amount'];

    /** @var \Drupal\commerce_easytransac\Plugin\Commerce\PaymentGateway\PayByBankInterface $payment_gateway_plugin */
    $payment_gateway_plugin = $this->plugin;
    $payment_gateway_plugin_config = $payment_gateway_plugin->getConfiguration();

    $order = $payment->getOrder();
    $customer = \Drupal::service('commerce_easytransac.easytransac')
      ->getCustomer($order);

    // Define the existing EasyTransac client ID when possible.
    if (!empty($customer_remote_id = $payment_gateway_plugin->getRemoteCustomerId($order->getCustomer()))) {
      $customer->setClientId($customer_remote_id);
    }

    $amount = \Drupal::service('commerce_price.minor_units_converter')
      ->toMinorUnits($payment->getAmount());

    $transaction = (new PaymentPageTransaction())
      ->setAmount($amount)
      ->setClientIP($order->getIpAddress())
      ->setDescription((string) $this->t('Payment of order #@id', ['@id' => $order->id()]))
      ->setPreAuth(FALSE)
      ->setCustomer($customer)
      ->setOrderId($order->id())
      ->setLanguage($payment_gateway_plugin->getApplication()->getLanguage());

    // Only authorization will be done, no capture yet.
    if ($values['capture'] === FALSE) {
      $transaction
        ->setPreAuth(TRUE)
        ->setPreAuthDuration(!empty($payment_gateway_plugin_config['authorization_duration']) ? $payment_gateway_plugin_config['authorization_duration'] : EasyTransac::EASYTRANSAC_MIN_AUTH_DURATION);
    }

    $service = $payment_gateway_plugin->getClient();
    // Add a modifier to force the "SaveCard" flag to "yes".
    $service->setModifier(new Modifier());
    $response = (new PaymentPage())->execute($transaction);
    if (!$response->isSuccess()) {
      throw new PaymentGatewayException($response->getErrorMessage());
    }

    // Save the request identifier in the order for reference.
    /** @var \EasyTransac\Entities\PaymentPageInfos $payment_page_information */
    $payment_page_information = $response->getContent();
    $payment->set('request_id', $payment_page_information->getRequestId());
    $payment->save();
  }

}
